######################################################################
#                               cegma                                #
######################################################################
#
#     Core Eukaryotic Gene Mapping Approach.
#
#     Copyright (C) 2006 -          Genis Parra
#                                   Keith Bradnam
#                                     Ian Korf
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
######################################################################

# 	$Id: cegma.pl,v 1.6 2007/11/14 04:54:26 genis Exp $	

use strict;
use Getopt::Long;

##########################################################################
##                                                                      ##
##                     VARIABLES  AND DEFINITIONS                       ##
##                                                                      ##
##########################################################################

# MAIN VARIABLES 
my $PROGRAM = "cegma";
my $VERSION = "2.0";
 
my $CEGMA     = defined($ENV{'CEGMA'}) ? $ENV{'CEGMA'} : '.';
my $BIN       = defined($ENV{'CEGMA'}) ? "$CEGMA/bin" : '.';
my $CEGMAdata = defined($ENV{'CEGMA'}) ? "$CEGMA/data" : "../data";
my $TMP       = defined($ENV{'CEGMATMP'}) ? $ENV{'CEGMATMP'} : '/tmp';
my $TMPROOT   = "cegma_$$";
my $CEGMATMP  = "$TMP/$TMPROOT";

# PROGRAMS DEFINITION 

my $PARSEBLAST = "parseblast";
my $verbose_str = "";

##########################################################################
##                                                                      ##
##                          READING OPTIONS                             ##
##                                                                      ##
##########################################################################

# print help and exit if any option is given
&prt_Help 
	unless scalar(@ARGV); 

# getOptions Variables
my ( $genome, $protein, $output, $geneid_opt, $geneid_param, 
     $blast_opt, $score_cutoff, $gnw_file, $tbx,
     $blastdb, $annotation, $ofn, $ps_output, $max_intron, 
     $savefiles_flg, $verbose_flg, $ext_flg, 
     $mam_flg, $vrt_flg, $quiet_flg, $temp_flg, $help_flg,
     $interlen, $boundaries
     ) = ( undef, undef, undef, undef, undef, undef, undef,
           undef, undef, undef, undef, undef, undef, undef,
           0, 0, 0, 0, 0, 0, 0, 0, 5000, 2000 );

# file flags
my ($blast_flg, $hsps_flg, $geneid_flg, $plots_flg) = (1, 1, 1, 1);

# processes
my $status;

# reading options
&Which_Options();


##########################################################################
##                                                                      ##
##                          INITIAL CHECKINGS                           ##
##                                                                      ##
##########################################################################

# Cheking if the external programs are in the path.
system("which tblastn > /dev/null;") == 0 || do {
   &go_to_die("BLAST is not found or not executable")
};
system("which genewise > /dev/null;") == 0 || do {
   &go_to_die("genewise is not found or not executable")
};
system("which geneid > /dev/null;") == 0 || do {
   &go_to_die("geneid is not found or not executable")
};
system("which hmmsearch > /dev/null;") == 0 || do {
   &go_to_die("HMMER package is not found or not executable")
};


# Checking if all programs exists and are executable
(-e "$BIN/$PARSEBLAST" && -x _) || do {
   &go_to_die("$PARSEBLAST is not found or executable in $BIN")
};



##########################################################################
##                                                                      ##
##                          COMPUTING DATA                              ##
##                                                                      ##
##########################################################################

## Running genome_map.pl

&prt_Header("RUNNING GENOME INITIAL MAPPING")
    unless (defined($gnw_file));

&run_genome_map($genome,$protein,$tbx,$blastdb,$annotation,$interlen,$boundaries)
    unless (defined($gnw_file));

(!(-z "genome.chunks.fa")) || do {
   &go_to_die("No orthologous genes found in the first round.\n\n   blast didn't give any suitable aligment. \n\n   Use --ext option to check blast alignments (genome.blast file).")
};

## Running  local_map.pl
&prt_Header("RUNNING LOCAL MAPPING");

&run_local_map("local","-f")
    unless (defined($gnw_file));

&run_local_map("local","-g $gnw_file -f")
    if (defined($gnw_file));

(!(-z "local.genewise.gff")) || do {
   &go_to_die("No orthologous genes found in the first round.\n\n   genewise didn't give any suitable aligment.")
};

## Filtering the set of predicted proteins 
&prt_Header("FILTERING INITIAL PROTEINS");

system("$BIN/hmm_select $CEGMA/data/hmm_profiles/ local.geneid.fa $CEGMA/data/profiles_cutoff.tbl >  local.geneid.selected.fa");

system("grep \">\" local.geneid.selected.fa | sed 's/_.*\$//;s/>//' >  local.geneid.selected.id");

system("$BIN/FastaToTbl  genome.chunks.fa | grep -wf local.geneid.selected.id  | $BIN/TblToFasta > local.geneid.selected.dna ");

system ("grep -wf local.geneid.selected.id local.geneid.gff > local.geneid.selected.gff");

(!(-z "local.geneid.selected.id")) || do {
   &go_to_die("No orthologous genes found in the first round.\n\n     Check if blast, genewise and hmmer are properly installed and correspond to compatible versions.\n\n   Use --ext option to check intermediate files.")
};


## Filtering the set of predicted proteins 
&prt_Header("SELF TRAINING");

system ("mkdir geneid_params");

system ("$BIN/geneid-train local.geneid.selected.gff local.geneid.selected.dna geneid_params")  == 0 || do {
   &go_to_die("geneid-train did not work properly. Check PERL5LIB variable.")
};;

if ($max_intron) {
     open(INTRON, "> geneid_params/intron.max");
     print INTRON "First+:Internal+             Internal+:Terminal+          30:$max_intron block\n";
     close(INTRON);
}

system ("$CEGMA/bin/make_paramfile $CEGMA/data/self.param.template \\
  geneid_params/coding.initial.5.logs geneid_params/coding.transition.5.logs \\
  geneid_params/start.logs geneid_params/acc.logs geneid_params/don.logs \\
  geneid_params/intron.max >  geneid_params/self.param") == 0 || do {
   &go_to_die("make_paramfile did not work properly.")
};;

## Filtering the set of predicted proteins 
&prt_Header("ACCURATE LOCAL MAPPING");

&run_local_map("local_self","-g local.genewise.gff -d geneid_params/self.param");

## Filtering the set of predicted proteins 
&prt_Header("FINAL FILTERING");

system ("$BIN/hmm_select $CEGMA/data/hmm_profiles/ local_self.geneid.fa \\
     $CEGMA/data/profiles_cutoff.tbl \\
  >  $output.cegma.fa");


system ("grep \">\" $output.cegma.fa | sed 's/_.*\$//;s/>//' \\
  >  $output.cegma.id");

system ("$BIN/FastaToTbl  genome.chunks.fa | grep -wf $output.cegma.id \\
  | $BIN/TblToFasta > $output.cegma.local.dna ");

system ("grep -wf $output.cegma.id local_self.geneid.gff | sed 's/geneid_v.../cegma/'  \\
  > $output.cegma.local.gff ");

&convert_coordinates("genome.chunks.gff","$output.cegma.local.gff", "$output.cegma.gff");

system ("$BIN/completeness hmm_select.aln $CEGMA/data/completeness_cutoff.tbl \\
  > $output.completeness_report");

##########################################################################
##                                                                      ##
##                              ENDING                                  ##
##                                                                      ##
##########################################################################

## Removing extended files;
&clean_ext() unless $ext_flg;

## Removing temporary files
&clean_tmp() unless $temp_flg;

exit(0);

##########################################################################
##                                                                      ##
##                              THE END                                 ##
##                                                                      ##
##########################################################################
##########################################################################


##########################################################################
##                                                                      ##
##                            SUBROUTINES                               ##
##                                                                      ##
##########################################################################

# Checking fasta format sequence files
sub check_fasta_format() {
    my $file = $_[0];
    my ($n, $c) = (undef, 0);
    open(TMP,"< $file");
    while (<TMP>) {
        next unless /^>/;
        />(\S+)\b/ && do {
            $n = $1;
            $c++;
            next unless $c>1;
        };
        &go_to_die("FATAL ERROR !!! Multiple locus names found.\n  File \'$file\' must contain only one sequence definition.");
    }
    &go_to_die("FATAL ERROR !!! There is no '>' line, locus name not found.\n  Please, verify your fasta file \'$file\'") unless defined($n);
    return $n;
}

# Deleting extended files
sub clean_ext() {
    # Obtaining the list of extended files
    my @files = qw( extended.tbl genome.blast genome.blast.gff 
       genome.chunks.fa genome.chunks.gff genome.err local.err
       local.geneid local.geneid.fa local.geneid.gff 
       local.genewise local.genewise.gff local_self.err
       local_self.geneid local_self.geneid.fa local_self.geneid.gff
       local.geneid.selected.dna local.geneid.selected.fa 
       local.geneid.selected.gff local.geneid.selected.id hmm_select.aln
       geneid_params/acc.logs geneid_params/intron.max geneid_params/self.param
       geneid_params/start.logs geneid_params/don.logs 
       geneid_params/coding.initial.5.logs 
       geneid_params/coding.transition.5.logs  );

   # Unlinking the temporary files if they exist
    ($#files >= 0) and 
        ( unlink(@files)
          and &prt_Header("********* Extended files were deleted *********")
          or &prt_Header("********* Can't unlink @files : $! *********")
        ) or 
          &prt_Header("********* There are no extended files  *********");
    
    rmdir "geneid_params";

}

# Deleting temporary files on TMP
sub clean_tmp() {
    # Obtaining the list of temporary files
    opendir(DIR,"$TMP");
    my @files = map { "$TMP/$_" } grep { /^$TMPROOT/ } readdir(DIR);
    closedir(DIR);
    # Unlinking the temporary files if they exist
    ($#files >= 0) and 
        ( unlink(@files)
          and &prt_Header("********* Temporary files were deleted *********")
          or &prt_Header("********* Can't unlink @files : $! *********")
        ) or 
          &prt_Header("********* There are no temporary files in $TMP *********");
}

# Checking input sequence files
sub exists_file() {
    my @files = @_;
    my ($n, $r) = (' ', 0);
    foreach $n (@files) {
        $r++ if (-e "$n");
    };
    return $r;
}

# Get a fixed length string from a given string and filling char/s.
sub fill_mid() { 
   my $l = length($_[0]); 
   my $k = int(($_[1] - $l)/2); 
   return ($_[2] x $k).$_[0].($_[2] x ($_[1] - ($l+$k)));
}

# writing die messages to STDERR and clean_tmp before exit.
sub go_to_die() { 
   (print STDERR "\n   @_ \n\n") && &clean_tmp();
   &clean_ext() unless $ext_flg;
   exit(1) 
}

# section headers to STDERR
sub prt_Header() { 
   print STDERR ("*" x 80)."\n** ".&fill_mid("@_",74," ")." **\n".("*" x 80)."\n" unless $quiet_flg;
}
# print the command to be executed
sub prt_Command() { 
	print STDERR ("\n\n** ".&fill_mid("@_",74," ")." **\n\n");
}

# Run genome_map script
sub run_genome_map { 
    my ( $genome_file, $prot_file, $tblastn_file, $blastdb, 
	 $annotation, $interlen_val, $boundaries_val) = @_;
    my $genome_map;
    my $options = "";

    if ($tblastn_file){
	$options = " -b $tblastn_file ";
    }
    if ($blastdb){
	$options = $options . " -d $blastdb ";
    }
    if ($annotation){
	$options = $options . " -a $annotation ";
    }
    
    
    $genome_map="$BIN/genome_map -n genome -p 6 $options  -i $interlen_val -c $boundaries_val $prot_file $genome_file 2>genome.err";
    

    &prt_Command("$genome_map") if $verbose_flg;     

    $status = system($genome_map);
    if ($status != 0) {
	&go_to_die("FATAL ERROR !!! $? $!\n
          Unsuccessful genome_map command :\n  $genome_map")
	};   
}

sub run_local_map { 
    my ($outputfile,$option) = @_;

    my $local_map="$BIN/local_map -n $outputfile $option -h $CEGMA/data/hmm_profiles/ genome.chunks.fa 2>$outputfile.err";

    &prt_Command("$local_map") if $verbose_flg;     

    $status = system($local_map);
    if ($status != 0) {
	&go_to_die("FATAL ERROR !!! $? $!\n
          Unsuccessful local_map command :\n  $local_map")
	};   
}

 
sub run_parseblast { 
	my ($outputfile, $option) = @_;

	my $parse = "$BIN/$PARSEBLAST --bit-score --gff $option $verbose_str $tbx | gawk '{OFS=\"\\t\"; 
      print \$1,\$2,\$3,\$4,\$5,\$6,\$7,\$8,0}' > $outputfile;";
    &prt_Command("$parse") if $verbose_flg;     

	$status = system($parse);
	if ($status != 0) {
		&go_to_die("FATAL ERROR !!! $? $!\n
          Unsuccessful $PARSEBLAST command :\n  $parse")
	};   
}

# convert the coordinates of the fragments to the genomic ones
sub convert_coordinates() {
    my ($file1, $file2, $file3) = @_;
    my (%pos1, %pos2, %strand, %seq);
    

    open(FILE1,"< $file1");
    while (<FILE1>) {
	my @gff = split;
	$pos1{$gff[8]} = $gff[3];
	$pos2{$gff[8]} = $gff[4];
	$strand{$gff[8]} = $gff[6];
	$seq{$gff[8]} = $gff[0];
    }
    close(FILE1);

    open(FILE2,"< $file2");
    open(FILE3,"> $file3");
    while (<FILE2>) {
	my @gff = split;
	if ($gff[0] ne "#" && $strand{$gff[0]} eq "+"){
	    printf FILE3  "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n", $seq{$gff[0]}, $PROGRAM, $gff[2], 
	            $gff[3] + $pos1{$gff[0]} - 1, $gff[4] + $pos1{$gff[0]} - 1, 
                    $gff[5], "+", $gff[7], $gff[0];
	} elsif ($gff[0] ne "#" && $strand{$gff[0]} eq "-"){
	    printf FILE3  "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n", $seq{$gff[0]}, $PROGRAM, $gff[2], 
	            $pos2{$gff[0]} - $gff[3] + 1, $pos2{$gff[0]} - $gff[4] + 1, 
                    $gff[5], "-", $gff[7], $gff[0];
	} 
    }

    close(FILE2);
    close(FILE3);

    return(0);

}

# get a fixed length string from a given string and filling char/s.
sub Which_Options() {
    GetOptions( 
             "g|genome=s"  => \$genome       , # genome fasta file
             "p|protein=s" => \$protein      , # protein fasta file
	     "o|output=s"  => \$output       , # output file name
             "c=f"         => \$score_cutoff , # tblastx score cutoff
             "s|genewise=s"=> \$gnw_file     , # shrink hsp's by
             "t|tblastn=s" => \$tbx          , # read tblastn from file
             "d|blastdb=s" => \$blastdb      , # blastdb file
	     "a|annotation=s" => \$annotation, # annotation file
             "vrt"         => \$vrt_flg      , # vertebrate
             "mam"         => \$mam_flg      , # mamalian
	     "max_intron=f"=> \$max_intron   , # max intron for geneid
             "v|verbose"   => \$verbose_flg  , # verbose    
             "q|quiet"     => \$quiet_flg    , # quiet mode
             "ext"         => \$ext_flg      , # kepp extended files
             "tmp"         => \$temp_flg     , # keep temporary files
             "h|help|?"    => \$help_flg     , # print help
                ) || &prt_Help();
    &prt_Help if $help_flg;
    &prt_Header("Processing Command-Line Options");

   # do seq files exists
   &go_to_die("FATAL ERROR!!! --genome or --blastdb must be defined.") 
	   unless (defined($genome) || defined($blastdb));
   
   my $file_number = &exists_file($genome);
   &go_to_die("FATAL ERROR!!! Input file: \"$genome\" does not exist.")
	   unless $file_number == 1; 
     

   &go_to_die("FATAL ERROR!!! --verbose and --quiet are incompatible options.")
	   if ($verbose_flg && $quiet_flg);

   &go_to_die("FATAL ERROR!!! --intron_max should be bigger than 50 bp.")
	   if ($max_intron && $max_intron<50);

   $protein = "$CEGMA/data/kogs.fa"   if (!(defined($protein)));
   $output = "output" if (!(defined($output)));
   $tbx = 0   if (!(defined($tbx)));
   $annotation = 0   if (!(defined($annotation)));

#  $GeneidParam   = "$geneid_param" if defined($geneid_param);
   $verbose_str = " -v " if $verbose_flg;
   $interlen = "20000" if $vrt_flg;
   $boundaries = "5000" if $vrt_flg;
   $interlen = "40000" if $mam_flg;
   $boundaries = "5000" if $mam_flg;
   


} 

# prints help 
sub prt_Help() {
    open(HELP, "| cat") ;
    print HELP <<"+++EndOfHelp+++";
                                                                     $PROGRAM


PROGRAM:
                        $PROGRAM - $VERSION

                Core Eukaryotic Genes Mapping Approach

USAGE:
        
    $PROGRAM [options] <-g genomic_fasta_sequence> 

DESCRIPTION:

    CEGMA (Core Eukaryotic Genes Mapping Approach) is a pipeline for
    building a set of high reliable set of gene annotations in
    virtually any eukaryotic genome. It combines tblastn, genewise,
    hmmer, with geneid, an "ab initio" gene prediction program.

REQUIRES:

    $PROGRAM needs the pre-instalation of the following software:

     - geneid (geneid v 1.3.7)
     http://genome.imim.es/software/geneid/
     - genewise (wise2.2.3-rc7)
     http://www.ebi.ac.uk/Wise2/
     - hmmer (HMMER 2.3.2 [Oct 2003])
     http://hmmer.janelia.org/
     - wu-blast (TBLASTN 2.0MP-WashU [10-May-2005])
     http://blast.wustl.edu/
 
    Check that you have the right version of the previous software.
    If you are using a different version and experience any problem,
    please let us know !!!

ENVIRONMENT VARIABLES:

    You can specify the path where $PROGRAM can find the default files
    with the shell variable \"$PROGRAM\". 

    You also can specify the path for the tempotary files with the
    shell variable \"CEGMATMP\". Default value is /tmp.

	Setting those vars in Bourne-shell and C-shell:

     o Using a Bourne-Shell (e.g. bash):
           export CEGMA="path"
           export CEGMATMP="path"
           export PERL5LIB="\$PERL5LIB:\$CEGMA/lib" 

     o Using a C-Shell:
           setenv CEGMA "path"
           setenv CEGMATMP "path"
           setenv PERL5LIB "\$PERL5LIB:\$CEGMA/lib"
 
COMMAND-LINE OPTIONS:

	 Available options and a short description are listed here:

     -g, --genome      fasta file of the query sequence.

     -p, --protein     fasta file of the protein sequences.
                        (only used to run a subset of the 458 CEGs)

     -o, --output      ouput file prefix.

     -d, --blastdb     blast database for the genome sequence.

     -t, --tblastn     gff file with the blast results
                        (this option skips the blast step).

     -s, --genewise    gff file with the genewise alignment coordinates
                        (this options skips the blast step).

     -a, --annotation  gff file with gene annotations
                        (computes the coordinates for the annotations).

     --vrt             Optimization for vertebrate genomes
                        (intron size 20,000 bp).

     --mam             Optimization for mamalian genomes
                        (intron size 40,000 bp).

     --max_intron      Max intron size.

     -v, --verbose     Verbose mode, a report is sent to standard error 
                        (default is set to showing only WARNINGS).

     -q, --quiet       Quiet mode, do not show any message/warning
                       to standard error (only ERRORS are reported).

     -h, --help        Shows this help.

     --ext             Extended output with interemediated files.

     --tmp             Keep temporary files.

BUGS:    
  
    Report any problem to 'CEGMA\@spitfire.genomecenter.ucdavis'.

AUTHOR:  

    $PROGRAM has been developed by Genis Parra, Keith Bradnam 
    and Ian Korf.



GNU-GPL (C)                      November 2006                         $PROGRAM
+++EndOfHelp+++
    close(HELP);
    exit(1);
} # prt_help

