######################################################################
#                           genome_map                               #
######################################################################
#
#     cegma pipeline code
#
######################################################################
#      
# 	$Id: genome_map.pl,v 1.4 2008/01/17 19:31:15 genis Exp $	

use strict; use warnings; use sigtrap;
use FAlite;
use Getopt::Std;
use vars qw($opt_p $opt_n $opt_b $opt_d $opt_a $opt_i $opt_c);

##########################################################################
##                                                                      ##
##                     VARIABLES  AND DEFINITIONS                       ##
##                                                                      ##
##########################################################################
 
my $CEGMA = defined($ENV{'CEGMA'}) ? $ENV{'CEGMA'} : '.';
my $TMP   = defined($ENV{'CEGMATMP'}) ? $ENV{'CEGMATMP'} : '/tmp';

getopts('p:n:b:d:a:i:c:');

my $PREFIX = "";
my $PROTNUM;
my $BLAST_FILE_GFF;
my $ANNOT_FILE;
my $BLASTDB = "$TMP/genome$$.blastdb";
my $INTERLEN;
my $BOUNDARIES;


$PREFIX = $opt_n if $opt_n;
$PROTNUM = $opt_p if $opt_p;
$BLASTDB = $opt_d if $opt_d;
$BLAST_FILE_GFF = $opt_b if $opt_b;
$ANNOT_FILE = $opt_a if $opt_a;
$INTERLEN = $opt_i if $opt_i;
$BOUNDARIES = $opt_c if $opt_c;


die "

# This script takes a protien multifasta file and a genomic multifasta file
# and runs blast to get the best n candidates.

# If a blast file is provided it gets the data from there.

# we considered the sequence of the proteins from the same cluster to be 
# in  order.

genome_map.pl <protein_fasta_file> <genomic_fasta_file>
   -n <prefix of the output files>
   -p <number of species proteins per local proteins>
options:
   -b <blast file> 
   -d <genomic blast database>
   -a <annotation_gff>

" unless (@ARGV == 2 && $opt_p);


my $prot_file = $ARGV[0];
my $genome_file = $ARGV[1];

##########################################################################
##                                                                      ##
##                          INITIAL CHECKINGS                           ##
##                                                                      ##
##########################################################################

# Security check to not overwrite the results
if (!(-e "$prot_file" )) {
    print "File does not exist: $prot_file \n";
    exit(1);
}

if (!(-e "$genome_file" )) {
    print "File does not exist: $genome_file \n";
    exit(1);
}

#if (!(-e "$ANNOT_FILE" )) {
#    print "File does not exist: $ANNOT_FILE \n";
#    exit(1);
#}


if (-e "$PREFIX.blast" ) {
    print "File exist: $PREFIX.blast \n";
#    exit(1);
}

##########################################################################
##                                                                      ##
##                          COMPUTING DATA                              ##
##                                                                      ##
##########################################################################


# Building genomic blast database


system("xdformat -n -I -o $BLASTDB $genome_file")
 unless (defined($opt_d));

# Opening the protein fasta files
open(PROT_FILE, $prot_file) or die;
my $PROT = new FAlite(\*PROT_FILE);

# Main loop of execution
while (my $prot_entry = $PROT->nextEntry) {

    # Opening temporary fasta files
    open(TMP_PROT, "> $TMP/prot$$.fa") or die;
    print_fasta($prot_entry,\*TMP_PROT);
    my ($kog_id) = $prot_entry->def =~ /^>\S+___(\S+)/;

    for (my $c=1; $c<$PROTNUM; $c++){
	$prot_entry = $PROT->nextEntry;
	print_fasta($prot_entry,\*TMP_PROT);	
    }

    close(TMP_PROT);
    
    print STDERR "$kog_id \n";
    print "$kog_id \n";
    
    if (!($BLAST_FILE_GFF)) {
	# running blast
	system ("tblastn $BLASTDB $TMP/prot$$.fa \\
                -cpus 2 -hspmax 100 B=5 V=5 W=5 T=25 filter=seg lcfilter \\
                > $TMP/blast$$");

	
	system ("cat $TMP/blast$$ >> $PREFIX.blast");
	
	# filtering the result
	system ("$CEGMA/bin/parseblast -GSb $TMP/blast$$ | sort -k 1,2 -k 4n \\
             > $TMP/blast$$.gff ");
	system ("cat $TMP/blast$$.gff >> $PREFIX.blast.gff");
    } else {
	system ("grep $kog_id $BLAST_FILE_GFF > $TMP/blast$$.gff");
    }
    # obtaining the coordinates of the chunks 
    # and take the 5 best scoring ones.
    system ("$CEGMA/bin/get_chunks $INTERLEN $BOUNDARIES $TMP/blast$$.gff | sort -k 6rn | head -5 \\
               > $TMP/chunks$$.gff ");
    system ("cat $TMP/chunks$$.gff >> $PREFIX.chunks.gff");

     
    open (CHUNKS, "< $TMP/chunks$$.gff");
    my $xdget_option;

    while (<CHUNKS>) {
	my @gff = split;
	# Getting the genomic coordinates and converted to local
	# gawk script to get a reverse the coordinates
	if (defined ($ANNOT_FILE)){
	    system ("awk '\$1 == \"$gff[0]\" && \$4 > $gff[3] && \$5 < $gff[4]' \\
                      $ANNOT_FILE \\
                    | $CEGMA/bin/local_coord $gff[3] $gff[4] $gff[6] $gff[8] \\
                    >> $PREFIX.chunks.ann.gff");
	}

        if ($gff[6] eq "-"){
	    $xdget_option = "-r";
	}else{
	    $xdget_option = "";
	}
	# obtaining the genomic fragments
	system ("xdget -n  $xdget_option -a $gff[3] -b $gff[4] \\
                  $BLASTDB $gff[0] \\
	          | sed 's/>\\(.*\\)/>$gff[8] \\1/'>>  $PREFIX.chunks.fa;");
    } 
    close(CHUNKS);
}


close (PROT_FILE);

# removing temporary files

system("rm -f  $TMP/prot$$.fa $TMP/chunks$$.gff  $TMP/blast$$* $TMP/blast$$.gff $TMP/genome$$.blastdb");

exit(0);

##########################################################################
##                                                                      ##
##                              THE END                                 ##
##                                                                      ##
##########################################################################
##########################################################################


##########################################################################
##                                                                      ##
##                            SUBROUTINES                               ##
##                                                                      ##
##########################################################################



sub print_fasta {

    my ($entry, $file) = @_;
    my ($id, $seq) = split (/\n/,$entry);
    my ($i, $e) = (1, length($seq));
    print $file "$id\n";
    while ($i<=$e) {
	print  $file  substr($seq,$i-1,60)."\n"; 
	$i += 60;
    };
}
